const Promise = require('bluebird');
const { v4: uuidv4 } = require('uuid');

const { getContractAddressByNetworkName } = require('../util/contract-config');
const { registerDecider } = require('./validation-pools');
const { registerMatrixEventHandler, sendMatrixEvent, sendMatrixText } = require('../matrix-bot');
const { matrixPools, matrixUserToAuthorAddress, applicationData } = require('../util/db');
const {
  rollup, wallet, work2, dao,
} = require('../util/contracts');
const read = require('../util/forum/read');
const write = require('../util/forum/write');
const addPostWithRetry = require('../util/add-post-with-retry');

const {
  ETH_NETWORK,
} = process.env;

let batchWorker;

const rollupAddress = getContractAddressByNetworkName(ETH_NETWORK, 'Rollup');

const getBatchPostAuthorWeights = async (batchItems) => {
  const weights = {};
  await Promise.each(batchItems, async (postId) => {
    // TODO: try/catch
    const post = await read(postId);
    // TODO: try/catch
    const matrixPool = await matrixPools.get(postId);
    const { fee, result: { votePasses, quorumMet } } = matrixPool;
    post.authors.forEach(({ authorAddress, weightPPM }) => {
      weights[authorAddress] = weights[authorAddress] ?? 0;
      if (votePasses && quorumMet) {
        // scale by matrix pool outcome and strength
        weights[authorAddress] += weightPPM * fee;
      }
      // TODO: Rewards for policing
      // TODO: Propagation via references
    });
    // Rescale author weights so they sum to 1000000
    const sumOfWeights = Object.values(weights).reduce((t, v) => t + v, 0);
    const scaledWeights = weights
      .map((weight) => Math.floor((weight * 1000000) / sumOfWeights));
    const sumOfScaledWeights = Object.values(scaledWeights).reduce((t, v) => t + v, 0);
    scaledWeights[0] += 1000000 - sumOfScaledWeights;
    const authors = Object.entries(scaledWeights)
      .map(([authorAddress, weightPPM]) => ({ authorAddress, weightPPM }));
    return authors;
  });
};

const submitBatchPost = async (batchItems) => {
  const authors = await getBatchPostAuthorWeights(batchItems);
  // TODO: Compute citations as aggregate of the citations of posts in the batch
  const citations = [];
  const content = `Batch of ${batchItems.length} items`;
  const embeddedData = {
    matrixPools: batchItems.map((x) => x.postId),
    nonce: uuidv4().replace(/-/i, ''),
  };
  const sender = await wallet.getAddress();
  const contentToVerify = `${content}\n\n${JSON.stringify(embeddedData, null, 2)}`;
  const signature = await wallet.signMessage(contentToVerify);
  // Write to the forum database
  const { hash: batchPostId } = await write({
    sender, authors, citations, content, embeddedData, signature,
  });
  // Add rollup post on-chain
  await addPostWithRetry(authors, batchPostId, citations);
  // Call Rollup.submitBatch
  const poolDuration = 60;
  await rollup.submitBatch(batchPostId, batchItems.length, poolDuration);
};

const evaluateMatrixPoolOutcome = async (postId) => {
  const matrixPool = await matrixPools.get(postId);
  // This should already contain all the info we need to evaluate the outcome
  const { stakes, quorum, winRatio } = matrixPool;
  const stakedFor = stakes
    .filter((x) => x.inFavor)
    .reduce((total, { amount }) => total + amount, 0);
  const stakedAgainst = stakes
    .filter((x) => !x.inFavor)
    .reduce((total, { amount }) => total + amount, 0);
  const votePasses = stakedFor * winRatio[1] >= (stakedFor + stakedAgainst) * winRatio[0];
  const totalSupply = await dao.totalSupply();
  const quorumMet = BigInt(stakedFor + stakedAgainst) * quorum[1] >= totalSupply * quorum[0];
  const result = {
    stakedFor, stakedAgainst, totalSupply, votePasses, quorumMet,
  };
  console.log('Matrix pool outcome evaluated', result);
  matrixPool.result = result;
  await matrixPools.put(postId, matrixPool);
  sendMatrixEvent('io.dgov.pool.result', result);

  // Even if we're not the current batch worker, keep track of batch items
  let batchItems;
  try {
    batchItems = await applicationData.get('batchItems');
  } catch (e) {
    batchItems = [];
  }
  batchItems.push(postId);
  await applicationData.put('batchItems', batchItems);

  if (batchWorker === '0x0000000000000000000000000000000000000000') {
    // TODO: If there's no batch worker, we should stake our availability
    // and then submit the batch immediately.
    await submitBatchPost(batchItems);
  }

  if (batchWorker === await wallet.getAddress()) {
    // TODO: If we are the batch worker, we should wait an appropriate amout of time /
    // number of matrix pools before submitting a batch.
  }
};

const validateWorkEvidence = async (sender, post) => {
  let valid = false;
  if (sender === work2.target) {
    const expectedContent = 'This is a work evidence post';
    valid = post.content.startsWith(expectedContent);
  }
  console.log(`Work evidence ${valid ? 'matched' : 'did not match'} the expected content`);
  return valid;
};

const start = async () => {
  console.log('registering validation pool decider for rollup');
  registerDecider(async (pool, post) => {
    // If this is not sent by the work1 contract, it's not of interest here.
    if (pool.sender !== rollupAddress) return false;

    // A rollup post should contain
    // - a list of off-chain validation pools
    // - authorship corresponding to the result of those off-chain pools
    if (!post.embeddedData?.matrixPools) return false;

    // Our task here is to check whether the posted result agrees with our own computations
    const expectedAuthors = await getBatchPostAuthorWeights(post.embeddedData.matrixPools);
    if (expectedAuthors.length !== post.authors.length) return false;
    return post.authors.every(({ authorAddress, weightPPM }) => {
      const expectedAuthor = expectedAuthors.find((x) => x.authorAddress === authorAddress);
      return weightPPM === expectedAuthor.weightPPM;
    });
  });

  // Check for an assigned batch worker
  batchWorker = await rollup.batchWorker();
  console.log('At startup', { batchWorker });

  rollup.on('BatchWorkerAssigned', async (batchWorker_) => {
    console.log('Batch worker assigned:', batchWorker);
    batchWorker = batchWorker_;
  });

  /// `sender` is the address that called Rollup.addItem on chain, i.e. the Work2 contract.
  rollup.on('BatchItemAdded', async (postId, sender, fee) => {
    // If we are the batch worker or there is no batch worker, initiate a matrix pool
    if (batchWorker === await wallet.getAddress()
      || batchWorker === '0x0000000000000000000000000000000000000000') {
      const duration = 20;
      const quorum = [1, 3];
      const winRatio = [1, 2];
      const params = {
        sender, fee, duration, quorum, winRatio,
      };

      let post;
      try {
        post = await read(postId);
      } catch (e) {
        console.error(`Post ID ${postId} not found`);
        return;
      }

      // Initialize a matrix pool
      try {
        await matrixPools.get(postId);
        // If this doesn't throw, it means we or someone else already sent this event
        console.log(`Matrix pool start event has already been sent for postId ${postId}`);
      } catch (e) {
        if (e.status === 404) {
          console.log('sending matrix pool start event');
          const { roomId, eventId } = sendMatrixEvent('io.dgov.pool.start', {
            postId,
            ...params,
          });
          // Register our own stake and send a message
          const currentRep = await dao.balanceOf(await wallet.getAddress());
          const valid = await validateWorkEvidence(sender, post);
          const stake = { amount: currentRep, account: await wallet.getAddress(), inFavor: valid };
          sendMatrixEvent('io.dgov.pool.stake', { postId, amount: currentRep, inFavor: valid });
          await matrixPools.put(postId, {
            postId,
            roomId,
            eventId,
            ...params,
            stakes: [stake],
          });
        } else {
          throw e;
        }

        // Since we're assuming responsibility as the batch worker,
        // set a timeout to evaulate the outcome
        setTimeout(() => evaluateMatrixPoolOutcome(postId), duration * 1000);
      }
    }
  });

  registerMatrixEventHandler(async (client, roomId, event) => {
    switch (event.type) {
      case 'io.dgov.pool.start': {
        const { postId, sender, ...params } = event.content;
        // We can use LevelDB to store information about validation pools
        const eventId = event.event_id;
        console.log('Matrix pool started', { postId, ...params });
        // Validate the target post, and stake for/against
        let post;
        try {
          post = await read(postId);
        } catch (e) {
          console.error(`Post ID ${postId} not found`);
          break;
        }
        const currentRep = await dao.balanceOf(await wallet.getAddress());
        const valid = await validateWorkEvidence(sender, post);
        const stake = { amount: currentRep, account: await wallet.getAddress(), inFavor: valid };
        sendMatrixEvent('io.dgov.pool.stake', { postId, amount: currentRep, inFavor: valid });
        await matrixPools.put(postId, {
          roomId, eventId, ...params, stakes: [stake],
        });
        break;
      }
      case 'io.dgov.pool.stake': {
        const { postId, amount, inFavor } = event.content;
        let account;
        try {
          account = await matrixUserToAuthorAddress(event.sender);
        } catch (e) {
          // Error, sender has not registered their matrix identity
          sendMatrixText(`Matrix user ${event.sender} has not registered their wallet address`);
          return;
        }
        let matrixPool;
        try {
          matrixPool = await matrixPools.get(postId);
        } catch (e) {
          // Error. matrix pool not found
          sendMatrixText(`Received stake for unknown matrix pool, for post ${postId}`);
          return;
        }
        const stake = { account, amount, inFavor };
        matrixPool.stakes = matrixPool.stakes ?? [];
        matrixPool.stakes.push(stake);
        await matrixPools.put(postId, matrixPool);
        console.log(`registered stake in matrix pool for post ${postId} by ${account}`);
        break;
      }
      case 'io.dgov.pool.result': {
        // This should be sent by the current batch worker
        // TODO: Compare batch worker's result with ours to verify and provide early warning
        break;
      }
      case 'io.dgov.rollup.submit': {
        // This should include the identifier of the on-chain validation pool
        // TODO: Compare batch worker's result with ours to verify
        break;
      }
      default:
    }
  });

  // Put in our availability stake to do the rollup work
  // Then check if there is a rollup worker.
  // It's possible we're the first or only worker;
  // if there is no worker assigned yet, we can submit the first one.
  // The procedure will be the same every time we are the rollup worker:
  // - Stake availability to be the next rollup worker;
  // - Create a rollup post -- compute authorship based on matrix pool results
  // - Send matrix event 'io.dgov.rollup.submit'? May not be necessary,
  //   as there will be notification from the contract itself.
  //   However, it would be nice to have in the tiemline of the room,
  //   and it could be nice to give the participants a heads up to expect the batch.
  //   We could even require it as part of batch validation.
  // - Call DAO.addPost(authors, postId)
  // - Call Rollup.submitBatch(postId, batchSize, poolDuration)
};

module.exports = {
  start,
};
